use chrono::Days;
use chrono::NaiveDate;
use chrono::NaiveDateTime;
#[cfg(feature = "test-helpers")]
use fake::{Fake, Faker};
use sqlx::PgPool;

use crate::END_OF_DAY;
use crate::START_OF_DAY;
use crate::models::Assignment;
use crate::models::Availability;
use crate::models::Event;
use crate::models::Function;
use crate::models::Location;
use crate::models::Role;
use crate::models::User;
use crate::validation::AsyncValidate;
use crate::validation::AsyncValidateError;
use crate::validation::start_date_time_lies_before_end_date_time;

pub struct EventChangeset {
    pub time: (NaiveDateTime, NaiveDateTime),
    pub name: String,
    pub location_id: i32,
    pub voluntary_wachhabender: bool,
    pub voluntary_fuehrungsassistent: bool,
    pub amount_of_posten: i16,
    pub clothing: i32,
    pub note: Option<String>,
}

pub struct EventContext<'a> {
    pub pool: &'a PgPool,
    pub event: Option<i32>,
    pub user: &'a User,
}

impl<'a> AsyncValidate<'a> for EventChangeset {
    type Context = EventContext<'a>;

    async fn validate_with_context(
        &self,
        context: &'a Self::Context,
    ) -> Result<(), AsyncValidateError> {
        let Some(location) = Location::read_by_id(context.pool, self.location_id).await? else {
            return Err(AsyncValidateError::new(
                "Der angegebene Veranstaltungsort existiert nicht.",
            ));
        };

        user_is_admin_or_area_manager_of_event_location(context.user, &location)?;

        start_date_time_lies_before_end_date_time(&self.time.0, &self.time.1)?;

        let mut minimum_amount_of_posten = 0_i16;
        if let Some(id) = context.event {
            let event = Event::read_by_id_including_location(context.pool, id)
                .await?
                .unwrap();
            let assignments_for_event =
                Assignment::read_all_by_event(context.pool, event.id).await?;
            minimum_amount_of_posten = assignments_for_event
                .iter()
                .filter(|a| a.function == Function::Posten)
                .count() as i16;

            time_can_be_extended_if_edit(&self.time, &event, &assignments_for_event, context.pool)
                .await?;
            date_unchanged_if_edit(&self.time, &event.start.date())?;
            can_unset_wachhabender(&self.voluntary_wachhabender, &assignments_for_event)?;
            can_unset_fuehrungsassistent(
                &self.voluntary_fuehrungsassistent,
                &assignments_for_event,
            )?;
            if location.area_id != event.location.unwrap().area_id {
                return Err(AsyncValidateError::new(
                    "Veranstaltungsort kann nicht zu einem Ort außerhalb des initialen Bereichs geändert werden.",
                ));
            }
        }

        if !(minimum_amount_of_posten..=100).contains(&self.amount_of_posten) {
            return Err(AsyncValidateError::new(
                "Die Anzahl der Posten darf nicht kleiner als die Anzahl der bereits geplanten Posten und maximal 100 sein.",
            ));
        }

        Ok(())
    }
}

fn user_is_admin_or_area_manager_of_event_location(
    user: &User,
    location: &Location,
) -> Result<(), AsyncValidateError> {
    if user.role != Role::Admin
        && !(user.role == Role::AreaManager && user.area_id == location.area_id)
    {
        return Err(AsyncValidateError::new(
            "Du verfügst nicht über die Berechtigung, diese Veranstaltung zu erstellen bzw. zu bearbeiten.",
        ));
    }
    Ok(())
}

fn date_unchanged_if_edit(
    time: &(NaiveDateTime, NaiveDateTime),
    date_in_db: &NaiveDate,
) -> Result<(), AsyncValidateError> {
    if time.0.date() != *date_in_db {
        return Err(AsyncValidateError::new("event date can't be changed"));
    }

    Ok(())
}

async fn time_can_be_extended_if_edit(
    time: &(NaiveDateTime, NaiveDateTime),
    event: &Event,
    assignments_for_event: &Vec<Assignment>,
    pool: &PgPool,
) -> Result<(), AsyncValidateError> {
    let start = event.start.date();
    let end = event.start.date().checked_add_days(Days::new(1)).unwrap();
    let mut common_time = (start.and_time(START_OF_DAY), end.and_time(END_OF_DAY));

    for assignment in assignments_for_event {
        let availability = Availability::read_by_id(pool, assignment.availability_id)
            .await?
            .unwrap();
        let all_assignments =
            Assignment::read_all_by_availability(pool, assignment.availability_id).await?;

        if all_assignments.len() == 1 {
            if availability.start > common_time.0 {
                common_time.0 = availability.start;
            }

            if availability.end < common_time.1 {
                common_time.1 = availability.end;
            }
        } else {
            let mut slots = vec![(availability.start, availability.end)];
            for a in all_assignments
                .iter()
                .filter(|x| x.event_id != assignment.event_id)
            {
                let (fit, rest) = slots
                    .into_iter()
                    .partition(|s| s.0 >= a.start && s.1 <= a.end);
                slots = rest;
                let fit = fit.first().unwrap();

                if fit.0 != a.start {
                    slots.push((fit.0, a.start));
                }

                if fit.1 != a.end {
                    slots.push((a.end, fit.1));
                }
            }

            let slot = slots
                .into_iter()
                .find(|s| s.0 >= assignment.start && s.1 <= assignment.end)
                .unwrap();

            if slot.0 > common_time.0 {
                common_time.0 = slot.0;
            }

            if slot.1 < common_time.1 {
                common_time.1 = slot.1;
            }
        }
    }

    let old_start_time = common_time.0;
    let new_start_time = time.0;
    let old_end_time = common_time.1;
    let new_end_time = time.1;

    if new_start_time < old_start_time {
        return Err(AsyncValidateError::new(
            "starttime lies outside of available time for assigned people",
        ));
    }

    if new_end_time > old_end_time {
        return Err(AsyncValidateError::new(
            "endtime lies ouside of available time for assigned people",
        ));
    }

    Ok(())
}

fn can_unset_fuehrungsassistent(
    fuehrungsassistent_required: &bool,
    assignments_for_event: &Vec<Assignment>,
) -> Result<(), AsyncValidateError> {
    if !*fuehrungsassistent_required
        && assignments_for_event
            .iter()
            .any(|a| a.function == Function::Fuehrungsassistent)
    {
        return Err(AsyncValidateError::new(
            "fuehrungsassistent can't be set to not by ff, because a person is already assigned",
        ));
    }

    Ok(())
}

fn can_unset_wachhabender(
    voluntary_wachhabender: &bool,
    assignments_for_event: &Vec<Assignment>,
) -> Result<(), AsyncValidateError> {
    if !*voluntary_wachhabender
        && assignments_for_event
            .iter()
            .any(|a| a.function == Function::Wachhabender)
    {
        return Err(AsyncValidateError::new(
            "wachhabender can't be set to not by ff, because a person is already assigned",
        ));
    }

    Ok(())
}

#[cfg(feature = "test-helpers")]
impl EventChangeset {
    pub fn create_for_test(start: NaiveDateTime, end: NaiveDateTime) -> EventChangeset {
        let changeset = EventChangeset {
            time: (start, end),
            name: Faker.fake(),
            location_id: 1,
            voluntary_wachhabender: true,
            voluntary_fuehrungsassistent: true,
            amount_of_posten: 5,
            clothing: 1,
            note: None,
        };

        changeset
    }
}
